import argparse
import time
from joblib import Parallel, delayed



from noise import NoiseGenFactory
from utils import *
from amp import AMPALGO
from pca import get_pcapack
from empbayes import NonparBayes

m = 3000
n = 4000
niters = 5 # TODO 

ERRORMETRICS ={
    "mse": get_mse,
    "subspaceDist": get_subspace_distance
}

def multTrial(signals, noise_type, rep):
    if not isinstance(signals, list):
        signals = [signals]
    assert len(signals) == 2
    rank = len(signals)
    rank = len(signals)
    print("Noise type", noise_type) 
    NoiseGen = NoiseGenFactory.registry[noise_type]
    Qu, Ustar = simulate_structured_PC2(m, rank, reverse = True)
    Qv, Vstar = simulate_structured_PC2(n, rank, reverse = True)
    print("m, n = {} {}".format(m,n))
        
    fpattern = "../records/ERROR/PLACEHOLDER_n{}/{:.2f}_{:.2f}_DIRECTION.csv".format(noise_type, signals[0], signals[1]) 
    # if args.local:
    #     compareAlgos(signals, Qu, Ustar, Qv, Vstar, fpattern, NoiseGen, 0)
    #     quit()
    
    Parallel(n_jobs = num_cores)\
        (delayed(compareAlgos)(signals, Qu, Ustar, Qv, Vstar, fpattern, NoiseGen, i)
        for i in np.arange(rep)
        )
    


def compareAlgos(signals, Qu, Ustar, Qv, Vstar, fpattern, NoiseGen, rngseed):
    success = 0
    while success < 1:
        try: 
            rngseed = int(time.time()) + rngseed * 79
            rank = len(signals)
            ## --- Noise --- ##
            noise = NoiseGen(m,n, rngseed)
            ## --- Data --- ##
            Y = Ustar * signals @ Vstar.T / np.sqrt(m*n) + noise.W

            pcapack = get_pcapack(Y, rank, niters)
            
            ## --- AMP:PCA --- ##
            Qu = extended_redirect_pc(Ustar, pcapack.F, Qu)
            Ustar = redirect_pc(Ustar, pcapack.F)
            Qv = extended_redirect_pc(Vstar, pcapack.G, Qv) 
            Vstar = redirect_pc(Vstar, pcapack.G)
            
            udenoiser = NonparBayes(truePriorLoc=Qu, to_save=False, fig_prefix=NoiseGen.name + "_f")
            vdenoiser = NonparBayes(truePriorLoc=Qv, to_save=False, fig_prefix=NoiseGen.name + "_g")
            ## --- Denoise --- ##
            # for algoname in ["riunamp", "gamp", "riamp"]:
            # algonames = ["riamp", "rriamp", "riunamp", "ggamp", "sriamp"]
            algonames = ['sriamp', "ggamp", "riunamp"]
            # algonames = ['sriamp']
            # if args.local:
            #     algonames = ["riamp", "sriamp"]

            ## --- Denoise --- ##
            algonames = ["ggamp", "riunamp", "sriamp"]
            
            Ualgos = []
            Valgos = []
            for algoname in algonames:
                amp = AMPALGO[algoname]
                U, V = amp(pcapack, udenoiser=udenoiser, vdenoiser=vdenoiser, niters = niters)

                print("{} stops at U{} with seed {}".format(algoname, U.shape[-1], rngseed))

                Ualgos.append(U)
                Valgos.append(V)

            ## --- Record AMP --- ##

            for it, algoname in enumerate(algonames):
                ferrorPattern = fpattern.replace("PLACEHOLDER", algoname)
                U = Ualgos[it]; V = Valgos[it]

                # U
                for error, get_error in ERRORMETRICS.items():
                    res = []
                    for i in range(niters+1):
                        if i < U.shape[-1]:
                            ans = get_error(U[:,:,i], Ustar)
                            res.append(ans)   
                        else:
                            res.append(ans) 
                    fname = ferrorPattern.replace("ERROR", error)
                    fname = fname.replace("DIRECTION", "u")
                    write_line_to_file(fname, res)
                    print(fname)
                    print(res)    

                # V
                for error, get_error in ERRORMETRICS.items():
                    res = []
                    for i in range(niters+1):
                        if i < V.shape[-1]:
                            ans = get_error(V[:,:,i], Vstar)
                            res.append(ans)   
                        else:
                            res.append(ans) 
                    fname = ferrorPattern.replace("ERROR", error)
                    fname = fname.replace("DIRECTION", "v")
                    write_line_to_file(fname, res)
                    print(fname)
                    print(res)    

            ## --- Record PCA --- ##
            algoname = "pca"
            ferrorPattern = fpattern.replace("PLACEHOLDER", algoname)
            
            # U 
            for error, get_error in ERRORMETRICS.items():
                pcaerr = get_error(pcapack.F, Ustar)
                shrinkagepcaerr = get_error(pcapack.F*pcapack.uinnerprods, Ustar)
                res = [pcaerr, shrinkagepcaerr]
                fname = ferrorPattern.replace("ERROR", error)
                fname = fname.replace("DIRECTION", "u")
                write_line_to_file(fname, res)
                print(fname)
                print(res)    

            # V
            for error, get_error in ERRORMETRICS.items():
                pcaerr = get_error(pcapack.G, Vstar)
                shrinkagepcaerr = get_error(pcapack.G*pcapack.vinnerprods, Vstar)
                res = [pcaerr, shrinkagepcaerr]

                fname = ferrorPattern.replace("ERROR", error)
                fname = fname.replace("DIRECTION", "v")
                write_line_to_file(fname, res)
                print(fname)
                print(res)    
            
        except Exception as e:
            print("Error, resimulate\n{}".format(e))

            if args.local == True:
                success +=1

            continue
        else:
            success+=1


if __name__=="__main__":
    
    parser = argparse.ArgumentParser(prog = 'CompareAlgo')
    parser.add_argument("--noisetype", type=str)
    parser.add_argument("--rep", type=int)
    parser.add_argument("--local", action="store_true")
    parser.add_argument("--s1", nargs = "?", type=float, const = 2, default=2)
    parser.add_argument("--s2", nargs = "?", type=float, const = 1.5, default=1.5)
    
    args = parser.parse_args()
    signals = [args.s1, args.s2]
    
    if args.local:
        num_cores = 1
        m = 1500
        n = 2000
        TOSAVE = False
        multTrial(signals, args.noisetype, args.rep)
        # Parallel(n_jobs = num_cores)\
        #     (delayed(multiTest)(signals, NoiseModel, SeModel, args.separ, args.rngseed, 3)
        #     for signals in np.arange(1.1,1.3,0.1)
        #     )
        quit()

    TOSAVE = False
    num_cores = 10
    multTrial(signals, args.noisetype, args.rep)
