import argparse
import time

from joblib import Parallel, delayed

from noise import NoiseFactory
from utils import *
from amp import AMPALGO
from pca import get_emppcapack
from empbayes import PrettyNonparBayes as NonparBayes

n = 4000
niters = 5
num_cores = 10

ERRORMETRICS ={
    "mse": get_mse,
    "subspaceDist": get_subspace_distance
}

def multTrial(signals, noise_type, rep):
    if not isinstance(signals, list):
        signals = [signals]
    assert len(signals) == 2
    rank = len(signals)
    print("n", n)
    print("noise type", noise_type)
    NoiseModel = NoiseFactory.registry[noise_type]
    Qu, Ustar = simulate_structured_PC2(n, rank, reverse = True)
    fpattern = "../records/ERROR/PLACEHOLDER_n{}/{:.2f}_{:.2f}.csv".format(args.noisetype, signals[0], signals[1]) 
    Parallel(n_jobs = num_cores)\
        (delayed(compareAlgos)(signals, Qu, Ustar, fpattern, NoiseModel, i)
        for i in range(rep)
    )

def compareAlgos(signals, Qu, Ustar, fpattern, NoiseModel, seedroot):
    
    print("Working on rep {}".format(seedroot))
    success = 0
    
    while success < 1:
        try: 
            rank = 2 # Bad 
            ## --- Noise --- ##
            rngseed = int(time.time()% 100000) + seedroot
            noise = NoiseModel(n, seed = rngseed)
            ## --- Data --- ##
            W = noise.W

            Y = Ustar * signals @ Ustar.T / n + W

            ## --- AMP:PCA --- ##
            # thetas = np.array(signals)
            # pcapack = get_pcapack(Y, thetas, noise, niters)
            pcapack = get_emppcapack(Y, rank, niters)
            Qu = extended_redirect_pc(Ustar, pcapack.F, Qu)
            Ustar = redirect_pc(Ustar, pcapack.F)
            denoiser = NonparBayes(truePriorLoc=Qu, to_save=False, fig_prefix= "{}_f".format(noise.name))

            
            ## --- Denoise --- ##
            algonames = ["ggamp", "riunamp", "sriamp"]
            
            Ualgos = []
            for algoname in algonames:
            # for algoname in ["sriamp"]:
                ampalgo = AMPALGO[algoname]
                U, T = ampalgo(pcapack, denoiser=denoiser, niters = niters)

                U = U.reshape((-1, T + 1, rank))
                print("{} stops at U{} with seed {}".format(algoname, T, rngseed))
                
                Ualgos.append(U)

            
            ## --- Record AMP --- ##

            for it, algoname in enumerate(algonames):
                ferrorPattern = fpattern.replace("PLACEHOLDER", algoname)
                
                U = Ualgos[it]
                
                for error, get_error in ERRORMETRICS.items():
                    res = []
                    for i in range(niters+1):
                        if i < U.shape[1]:
                            ans = get_error(U[:,i,:], Ustar)
                            res.append(ans)   
                        else:
                            res.append(ans) 
                    
                    fname = ferrorPattern.replace("ERROR", error)
                    
                    ## --- Save results --- ##
                    write_line_to_file(fname, res)
                    print(fname)
                    print(res)
            
            ## --- Record PCA --- ##
            algoname = "pca"
            ferrorPattern = fpattern.replace("PLACEHOLDER", algoname)
            
            for error, get_error in ERRORMETRICS.items():
                pcaerr = get_error(pcapack.F, Ustar)
                shrinkagepcaerr = get_error(pcapack.F*pcapack.innerprods, Ustar)
                res = [pcaerr, shrinkagepcaerr]
                fname = ferrorPattern.replace("ERROR", error)
                
                ## --- Save results --- ##
                write_line_to_file(fname, res)
                print(fname)
                print(res)


        
        except Exception as e:
            print("Rep {} Error, Resimulate\n{}".format(seedroot, e))
            
            if args.local == True:
                success +=1

            continue
        else:
            success+=1  
                

if __name__=="__main__":
    
    parser = argparse.ArgumentParser(prog = 'CompareAlgo')
    parser.add_argument("--noisetype", type=str)
    parser.add_argument("--rep", type=int)
    parser.add_argument("--local", action="store_true")
    parser.add_argument("--s1", nargs = "?", type=float, const = 2.0, default=2.0)
    parser.add_argument("--s2", nargs = "?", type=float, const = 1.6, default=1.6)
    
    args = parser.parse_args()


    assert args.noisetype in NoiseFactory.registry

    signals = [args.s1, args.s2]
        
    if args.local:
        num_cores = 1
        n = 2000
        multTrial(signals, args.noisetype, 1)
        quit()

    multTrial(signals, args.noisetype, args.rep)

