import argparse
import time

from joblib import Parallel, delayed

from noise import NoiseFactory
from utils import *
from amp import AMPALGO
from pca import get_emppcapack
from empbayes import PrettyNonparBayes as NonparBayes

n = 4000
niters = 5
num_cores = 10


def multTrial(signals, noise_type, rep):
    if not isinstance(signals, list):
        signals = [signals]
    assert len(signals) == 2
    rank = len(signals)
    print("n", n)
    print("noise type", noise_type)
    NoiseModel = NoiseFactory.registry[noise_type]
    Qu, Ustar = simulate_structured_PC2(n, rank, reverse = True)
    fpattern = "../records/PLACEHOLDER_n{}/{:.2f}_{:.2f}.csv".format(args.noisetype, signals[0], signals[1]) 
    Parallel(n_jobs = num_cores)\
        (delayed(compareAlgos)(signals, Qu, Ustar, fpattern, NoiseModel, i)
        for i in range(rep)
    )

def compareAlgos(signals, Qu, Ustar, fpattern, NoiseModel, seedroot):
    print("Working on rep{}".format(seedroot))
    success = 0
    while success < 1:
        try: 
            print("rep {},another try".format(seedroot))
            rank = 2 # Bad 
            ## --- Noise --- ##
            rngseed = int(time.time()% 100000) + seedroot
            noise = NoiseModel(n, seed = rngseed)
            ## --- Data --- ##
            W = noise.W

            Y = Ustar * signals @ Ustar.T / n + W

            ## --- AMP:PCA --- ##
            # thetas = np.array(signals)
            # pcapack = get_pcapack(Y, thetas, noise, niters)
            pcapack = get_emppcapack(Y, rank, niters)
            Qu = extended_redirect_pc(Ustar, pcapack.F, Qu)
            denoiser = NonparBayes(truePriorLoc=Qu, to_save=False, fig_prefix= "{}_f".format(noise.name))

            ## --- Denoise --- ##
            for algoname in ["ggamp", "riunamp", "sriamp"]:
                denoiser = NonparBayes(truePriorLoc=Qu, to_save=True, fig_prefix= "{}_f".format(algoname))
                ampalgo = AMPALGO[algoname]
                U, T = ampalgo(pcapack, denoiser=denoiser, niters = niters)

                U = U.reshape((-1, T + 1, rank))
                res = []
                for i in range(niters+1):
                    if i < T+1:
                        ans = get_mse(U[:,i,:], Ustar)
                        res.append(ans)   
                    else:
                        res.append(ans) 

                fname = fpattern.replace("PLACEHOLDER", algoname)
                
                ## --- Save results --- ##
                write_line_to_file(fname, res)
                print(fname)
                print(res)
                print("Stops at U{}".format(T))
                print("RNGSEED = {}".format(rngseed))
        
        except Exception as e:
            print("Error, resimulate\n{}".format(e))
            continue
            
        else:
            success+=1        

if __name__=="__main__":
    
    parser = argparse.ArgumentParser(prog = 'CompareAlgo')
    parser.add_argument("--noisetype", type=str)
    parser.add_argument("--rep", type=int)
    parser.add_argument("--local", action="store_true")
    parser.add_argument("--s1", nargs = "?", type=float, const = 2.0, default=2.0)
    parser.add_argument("--s2", nargs = "?", type=float, const = 1.6, default=1.6)
    
    args = parser.parse_args()


    assert args.noisetype in NoiseFactory.registry

    signals = [args.s1, args.s2]
        
    if args.local:
        num_cores = 1
        n = 4000
        multTrial(signals, args.noisetype, 1)
        quit()

    multTrial(signals, args.noisetype, args.rep)

