'''
rad 

theta -> lambda
1 1.6
1.5 2
1.6 2.08
==========
uniform
1 1.8
1.4 2.05
1.5 2.11
'''

import argparse
import time

from joblib import Parallel, delayed

from noise import NoiseFactory
from utils import *
from amp import AMPALGO
from pca import get_emppcapack
from empbayes import PrettyNonparBayes as NonparBayes

n = 4000
niters = 5
num_cores = 10

ERRORMETRICS ={
    "mse": get_mse,
    "subspaceDist": get_subspace_distance
}


def multTrial(signals, noise_type, rep):
    if not isinstance(signals, list):
        signals = [signals]
    assert len(signals) == 1
    rank = len(signals)
    NoiseModel = NoiseFactory.registry[noise_type]
    Qu, Ustar = simulate_structured_PC(n, rank)
    fpattern = "../records/ERROR/rank1/PLACEHOLDER_n{}/{:.2f}.csv".format(noise_type, signals[0]) 
    print("I can here here with fpattern {}".format(fpattern))
    for i in range(rep):
        compareAlgos(signals, Qu, Ustar, fpattern, NoiseModel, i)


def compareAlgos(signals, Qu, Ustar, fpattern, NoiseModel, seedroot):
    print("Working on rep{}".format(seedroot))
    success = 0
    while success < 1:
        try: 
            print("rep {}, another try".format(seedroot))
            rank = len(signals) # Bad 
            ## --- Noise --- ##
            rngseed = int(time.time()) + seedroot * 97
            noise = NoiseModel(n, seed = rngseed)
            ## --- Data --- ##
            W = noise.W

            Y = Ustar * signals @ Ustar.T / n + W

            ## --- AMP:PCA --- ##
            # pcapack = get_pcapack(Y, thetas, noise, niters)
            pcapack = get_emppcapack(Y, rank, niters)
            Qu = extended_redirect_pc(Ustar, pcapack.F, Qu)
            Ustar = redirect_pc(Ustar, pcapack.F)

            ## --- Denoise --- ##
            if noise.name == "emprad":
                theta_threshold = 1.55
            if noise.name == "empuniform":
                theta_threshold = 1.35
            if noise.name == "centeredBeta":
                theta_threshold = 1.55
            if np.array(signals).min() > theta_threshold:
                print("I am here")
                algonames = ["ggamp", "riunamp", "sriamp"] 
            else:
                print("Not above phase transition")
                algonames = ["riunamp", "sriamp"]

            Ualgos = []
            for algoname in algonames:
                denoiser = NonparBayes(truePriorLoc=Qu, to_save=False, fig_prefix= "{}_f".format(algoname +  noise.name))
                ampalgo = AMPALGO[algoname]
                U, T = ampalgo(pcapack, denoiser=denoiser, niters = niters)

                U = U.reshape((-1, T + 1, rank))
                print("{} stops at U{} with seed {}".format(algoname, T, rngseed))

                Ualgos.append(U)

            ## --- Record AMP --- ##
            for it, algoname in enumerate(algonames):
                ferrorPattern = fpattern.replace("PLACEHOLDER", algoname)
                
                U = Ualgos[it]
                
                for error, get_error in ERRORMETRICS.items():
                    res = []
                    for i in range(niters+1):
                        if i < U.shape[1]:
                            ans = get_error(U[:,i,:], Ustar)
                            res.append(ans)   
                        else:
                            res.append(ans) 
                    
                    fname = ferrorPattern.replace("ERROR", error)
                    
                    ## --- Save results --- ##
                    write_line_to_file(fname, res)
                    print(fname)
                    print(res)

            ## --- Record PCA --- ##
            algoname = "pca"
            ferrorPattern = fpattern.replace("PLACEHOLDER", algoname)
            
            for error, get_error in ERRORMETRICS.items():
                pcaerr = get_error(pcapack.F, Ustar)
                shrinkagepcaerr = get_error(pcapack.F*pcapack.innerprods, Ustar)
                res = [pcaerr, shrinkagepcaerr]
                fname = ferrorPattern.replace("ERROR", error)
                
                ## --- Save results --- ##
                write_line_to_file(fname, res)
                print(fname)
                print(res)

        except Exception as e:
            print("Error, resimulate\n{}".format(e))
            continue
        else:
            success+=1        

if __name__=="__main__":
    
    parser = argparse.ArgumentParser(prog = 'VaryTheta')
    parser.add_argument("--noisetype", type=str)
    parser.add_argument("--rep", type=int)
    parser.add_argument("--local", action="store_true")

    args = parser.parse_args()

    assert args.noisetype in ["empgoe", "empuniform", "emprad", "centeredBeta", "rad"]

    rep = args.rep
    noise_type = args.noisetype

    if args.local:
        num_cores = 1
        n = 1000
        rep = 1 
        for signals in np.arange(0.4,1,1): 
            multTrial(signals, noise_type, rep)
            # Parallel(n_jobs = num_cores)\
            #     (delayed(multTrial)(signals, noise_type, rep)
            #     for signals in np.arange(0.1,1,1)
            # )    
        quit()

    num_cores = 10
    Parallel(n_jobs = num_cores)\
        (delayed(multTrial)(signals, noise_type, rep)
        for signals in np.flip(np.arange(0.1,2.1,0.1))
    )    
