
import os
def mkdir(path):
    if not os.path.exists(path):
        os.makedirs(path)   

mkdir("../records")
mkdir("../figures")

import numpy as np
def redirect_pc(pc, reference):
    '''sample usage see tutorial.html
    we must redirect the truth
    redirect_pc(ustar, pcapack.U)
    '''
    U = pc 
    Ustar = reference
    M = U.T @ Ustar
    s = np.ones(len(M))
    for i in range(len(M)):
        if M[i,i] < 0:
            s[i] = -1
    return U * s

def extended_redirect_pc(pc, reference, pcLocs):
    '''sample usage see tutorial.html
    we must redirect the truth
    redirect_pc(ustar, pcapack.U)
    '''
    U = pc 
    Ustar = reference

    M = U.T @ Ustar
    s = np.ones(len(M))
    for i in range(len(M)):
        if M[i,i] < 0:
            s[i] = -1
    return pcLocs * s

def normalize_pc(U):
    return U/np.sqrt((U**2).sum(axis = 0)) * np.sqrt(len(U))

def simulate_structured_PC(n, k=2):
    if k == 2:
        Qu = np.array([[1,0],[1,0],[-1,2], [-1, -2]])
        Qu = normalize_pc(Qu)
        # Qu, _ = np.linalg.qr(U)
        Ustar = np.repeat(Qu, int(n/4), axis = 0)
    if k == 1:
        Qu = np.array([-1,1])
        Ustar = np.repeat([-1,1], int(n/2))
    return Qu.reshape((-1,k)), Ustar.reshape((-1,k))

# def simulate_structured_PC2(n, k =2, reverse = False):
#     if k == 2:
#         Qu = np.array([[1,0],[1,0],[-1,2], [-1, -2]])
#         if reverse:
#             Qu = np.array([[0,1],[0,1],[2,-1], [-2, -1]])
#         # Qu = normalize_pc(Qu)
#         Ustar = np.repeat(Qu, int(n/4), axis = 0) + np.random.normal(0, 0.5, size = (n,k))
#         Ustar = normalize_pc(Ustar)
#     return Ustar, Ustar

def simulate_structured_PC2(n, k =2, reverse = False):
    if k == 2:
        Qu = np.array([[1,0],[1,0],[-1,2], [-1, -2]])
        if reverse:
            Qu = np.array([[0,1],[0,1],[2,-1], [-2, -1]])
        Qu = normalize_pc(Qu)
        Ustar = np.repeat(Qu, int(n/4), axis = 0)
    return Qu, Ustar


def get_mse(U,V):
    assert U.shape == V.shape
    n = U.shape[0]
    return np.square(U-V).sum()/n

def get_subspace_distance(U,V):
    assert U.shape == V.shape
    Qu, _ = np.linalg.qr(U, mode = "reduced")
    Qv, _ = np.linalg.qr(V, mode = "reduced")
    try: 
        _, s, _ = np.linalg.svd(Qu.T.dot(Qv))
    except:
        print("QuQv", Qu.T.dot(Qv))
        s = 1
    return np.sqrt(1 - min(np.min(s)**2, 1))


def get_avgfro_norm(A):
    return np.sqrt(np.square(A).mean())


def write_to_record(signal_strength, algo, noise_type, aligns):
    fpath = "records/" + algo + noise_type + str(signal_strength) + ".csv"
    aligns =  ",".join(map(str, aligns))
    with open(fpath, 'a') as the_file:
        the_file.write(aligns + '\n')

def write_line_to_file(fname, values):
    values =  ",".join(map(str, values))
    try:
        with open(fname, 'a') as the_file:
            the_file.write(values + '\n')
    except FileNotFoundError:
        fpath = os.path.split(fname)[0]
        mkdir(fpath)
        with open(fname, 'a') as the_file:
            the_file.write(values + '\n')



def write_to_recordPhaseTransition(algo, noise_type, aligns):
    fpath = "records/" + algo + noise_type + ".csv"
    aligns =  ",".join(map(str, aligns))
    with open(fpath, 'a') as the_file:
        the_file.write(aligns + '\n')

import pandas as pd
class Radiologist():

    def __init__(self, dpath, fname, **kwargs):
        self.fname = fname
        mkdir(dpath)
        self.fpath = dpath + fname + ".csv"
        self.expfields = kwargs.keys()
        expvals = kwargs.values()
        self.prefix = ",".join(map(str, expvals))
        self.case = {
            "mu": [],
            "cov": [],
            "cond": [],
            "invCov": [],
            "dist" : []
        }

    def record(self, kw, value, t):
        assert len(self.case[kw]) == t - 1
        self.case[kw].append(value)
        print(kw, value)
    
    def dump(self):
        with open(self.fpath, 'a') as the_file:
            for kw in self.case.keys():
                line = [self.prefix]
                line.append(kw)
                line += self.case[kw]
                line = ",".join(map(str, line))
                the_file.write(line + '\n')

    def parse(self):
        df = pd.read_csv(self.fpath, header = None)
        nrows, ncols = df.shape
        colnames = list(self.expfields) + ["indicator"] + list(map(str, range(ncols - len(self.expfields) - 1)))
        df.columns = colnames
        return df



    






